@@ -43,7 +43,7 @@ sub pre_bind
                 $self->log(level => 'err', message => "bind rewrite \"${$binddnp}\" failed");
                 return -1;
             }
-            ${$binddnp} = $str;
+            (${$binddnp}) = split(/\n/, $str);
         }
     }
 
@@ -77,14 +77,14 @@ sub pre_compare
                 $self->log(level => 'err', message => "compare rewrite \"${$dnp}\" failed");
                 return -1;
             }
-            ${$dnp} = $str;
+            (${$dnp}) = split(/\n/, $str);
 
             $str = $self->_rewriteParse($rule->{match}, $substitution, ${$avaStrp}, \%rwcache);
             if (!$str) {
                 $self->log(level => 'err', message => "compcare rewrite \"${$dnp}\" failed");
                 return -1;
              }
-             ${$avaStrp} = $str;
+             (${$avaStrp}) = split(/\n/, $str);
         }
     }
 
@@ -118,14 +118,28 @@ sub pre_search
                 $self->log(level => 'err', message => "search rewrite \"${$basep}\" failed");
                 return -1;
             }
-            ${$basep} = $str;
+            (${$basep}) = split(/\n/, $str);
 
-            $str = $self->_rewriteParse($rule->{match}, $substitution, ${$filterStrp}, \%rwcache);
-            if (!defined($str)) {
-                $self->log(level => 'err', message => "search rewrite \"${$filterStrp}\" failed");
-                return -1;
+            my @elts = (${$filterStrp} =~ /\(([^()]+)\)/g);
+            for (my $i = 0; $i < @elts; $i++) {
+                $str = $self->_rewriteParse($rule->{match}, $substitution, $elts[$i], \%rwcache);
+                if (!defined($str)) {
+                    $self->log(level => 'err', message => "search rewrite \"${$filterStrp}\" failed");
+                   return -1;
+                }
+
+                my $elt;
+                foreach my $line (split(/\n/, $str)) {
+                    if ($elt) {
+                        $elt = "(&$elt($line))";
+                    } else {
+                        $elt = "($line)";
+                    }
+                }
+
+                $elts[$i] =~ s/([.*+?\[\]()|\^\$\\])/\\$1/g;
+                ${$filterStrp} =~ s/\($elts[$i]\)/$elt/;
             }
-            ${$filterStrp} = $str;
         }
     }
 
@@ -146,6 +160,7 @@ sub post_search
 
     foreach my $rule (@{$conf->{rewrite}}) {
         if ($rule->{context} eq 'searchResult') {
+            my %rwcache;
             for (my $i = 0; $i < @{$entriesp}; $i++) {
                 my $entryStr = encode('utf8', ${$entriesp}[$i]);
                 my (@line) = split(/\n/, $entryStr);
@@ -154,9 +169,7 @@ sub post_search
                     next;
                 }
 
-                my %rwcache;
                 my $rc = 0;
-
                 my $substitution = $rule->{substitution};
                 $substitution =~ s/%0/$entryStr/;
 
@@ -175,7 +188,7 @@ sub post_search
                         $rc = -1;
                         last;
                     } elsif ($line[$j]) {
-                        push(@replaced, $line[$j]);
+                        push(@replaced, split(/\n/, $line[$j]));
                     }
                 }
                 if (!$rc) {
@@ -218,7 +231,7 @@ sub pre_modify
                 $self->log(level => 'err', message => "modify rewrite \"${$dnp}\" failed");
                 return -1;
             }
-            ${$dnp} = $str;
+            (${$dnp}) = split(/\n/, $str);
 
             my @mod_list;
             while (@{$listp} > 0) {
@@ -292,7 +305,7 @@ sub pre_add
                 $self->log(level => 'err', message => "add rewrite \"${$dnp}\" failed");
                 return -1;
             }
-            ${$dnp} = $str;
+            (${$dnp}) = split(/\n/, $str);
 
             my (@line) = split(/\n/, ${$entryStrp}[0]);
 
@@ -337,7 +350,7 @@ sub pre_modrdn
                 $self->log(level => 'err', message => "modrdn rewrite \"${$dnp}\" failed");
                 return -1;
             }
-            ${$dnp} = $str;
+            (${$dnp}) = split(/\n/, $str);
 
             $str = $self->_rewriteParse($rule->{match}, $substitution, ${$argsp}[0], \%rwcache);
             if (!$str) {
@@ -371,15 +384,16 @@ sub pre_delete
                 next;
             }
 
+            my %rwcache;
             my $substitution = $rule->{substitution};
             $substitution =~ s/%0/${$dnp}/;
 
-            my $str = $self->_rewriteParse($rule->{match}, $substitution, ${$dnp});
+            my $str = $self->_rewriteParse($rule->{match}, $substitution, ${$dnp}, \%rwcache);
             if (!$str) {
                 $self->log(level => 'err', message => "delete rewrite \"${$dnp}\" failed");
                 return -1;
             }
-            ${$dnp} = $str;
+            (${$dnp}) = split(/\n/, $str);
         }
     }
 
@@ -433,8 +447,12 @@ sub _rewriteParse
 {
     my $self = shift;
     my ($match, $substitution, $str, $rwcache) = @_;
+    my $newstr;
 
     my @matches = ($str =~ /$match/gi);
+    if (!@matches) {
+         return $str;
+    }
 
     # replace variables
     for (my $i = 0; $i < @matches; $i++) {
@@ -442,10 +460,12 @@ sub _rewriteParse
         $substitution =~ s/%$num/$matches[$i]/g;
     }
 
-    if ($str =~ /$match/i) {
+    foreach my $substline ($self->_splitSubst($substitution)) {
+        my $oldstr = $str;
+
         # do functions
-        my @subs = ($substitution);
-        my @rwmaps = ($substitution =~ /%{([^(]*\([^)]*\).*)}/g);
+        my @substs = ($substline);
+        my @rwmaps = ($substline =~ /%{([^(]*\([^)]*\).*)}/g);
         foreach my $rwmap (@rwmaps) {
             my @values;
             my $key = lc($rwmap);
@@ -470,28 +490,57 @@ sub _rewriteParse
             }
 
             $rwmap =~ s/([.*+?\[\]()|\^\$\\])/\\$1/g;
-            my @tmpsubs;
-            foreach my $sub (@subs) {
+            my @tmpsubsts;
+            foreach my $subst (@substs) {
                 foreach my $value (@values) {
-                    my $tmpsub = $sub;
-                    $tmpsub =~ s/%{$rwmap}/$value/;
-                    push(@tmpsubs, $tmpsub);
+                    my $tmpsubst = $subst;
+                    $tmpsubst =~ s/%{$rwmap}/$value/;
+                    push(@tmpsubsts, $tmpsubst);
                 }
             }
-            undef(@subs);
-            @subs = @tmpsubs;
+            undef(@substs);
+            @substs = @tmpsubsts;
         }
 
         my @strs;
-        foreach my $sub (@subs) {
-            my $tmpstr = $str;
-            $tmpstr =~ s/$match/$sub/gi;
+        foreach my $subst (@substs) {
+            my $tmpstr = $oldstr;
+            $tmpstr =~ s/$match/$subst/gi;
             push(@strs, $tmpstr);
         }
-        $str = join("\n", @strs);
+        if ($newstr) {
+            $newstr = "$newstr\n".join("\n", @strs);
+        } else {
+            $newstr = join("\n", @strs);
+        }
     }
 
-    return $str;
+    return $newstr;
+}
+
+sub _splitSubst
+{
+    my $self = shift;
+    my ($substitution) = @_;
+    my @substs;
+
+    my $prevpos = 0;
+    my $oldpos = 0;
+    while ((my $pos = index($substitution, "\n", $oldpos)) > 0) {
+        my $str = substr($substitution, $prevpos, $pos - $prevpos);
+        if (index($str, "%{", $oldpos - $prevpos) > 0) {
+            if (index($str, "(", $oldpos - $prevpos) > 0) {
+                 $oldpos = index($substitution, ")", $oldpos);
+            }
+            $oldpos = index($substitution, "}", $oldpos);
+            next;
+        }
+        push(@substs, $str);
+        $prevpos = $oldpos = $pos + 1;
+    }
+    push(@substs, substr($substitution, $prevpos));
+
+    return @substs;
 }
 
 sub _rewriteMap
