@@ -128,7 +128,7 @@ sub bind
         return LDAP_SERVER_DOWN;
     }
 
-    my ($rc, $obj, $pkey) = $self->_getObject($binddn);
+    my ($rc, $obj, $pkeys) = $self->_getObject($binddn);
     if ($rc) {
         return $rc;
     }
@@ -137,7 +137,7 @@ sub bind
         my $entry;
         my $key;
 
-        ($rc, $key, $entry) = $self->_baseSearch($obj, $pkey, $binddn, 0, 0, 1, 0, undef);
+        ($rc, $key, $entry) = $self->_baseSearch($obj, $pkeys, $binddn, 0, 0, 1, 0, undef);
         if ($rc) {
             last DO;
         }
@@ -208,19 +208,24 @@ sub search
                 push (@match_entries, $entry);
                 $sizeLim--;
             }
-        } else {
-            $scope = 0;
         }
     }
 
     foreach my $objinfo (@objs) {
-        my ($obj, $pkey) = @{$objinfo};
+        my ($obj, $pkeys) = @{$objinfo};
         my $objbase = $base =~ /^$self->{suffix}$/ ? $obj->{suffix} : $base;
+        my $objscope;
         my $entry;
         my $key;
 
+        if ($scope == 1 && $base =~ /^$self->{suffix}$/ && $obj->{entrystr}) {
+            $objscope = 0;
+        } else {
+            $objscope = $scope;
+        }
+
         # search base entry
-        ($rc, $key, $entry) = $self->_baseSearch($obj, $pkey, $objbase, $scope, $deref, $sizeLim, $timeLim, $filter);
+        ($rc, $key, $entry) = $self->_baseSearch($obj, $pkeys, $objbase, $objscope, $deref, $sizeLim, $timeLim, $filter);
         if ($entry) {
             push(@match_entries, $entry);
             $sizeLim = $sizeLim - $entry;
@@ -234,7 +239,7 @@ sub search
         my @entries;
         my $keys;
 
-        ($rc, $keys, @entries) = $self->_childSearch($obj, $key, $objbase, $scope, $deref, $sizeLim, $timeLim, $filter);
+        ($rc, $keys, @entries) = $self->_childSearch($obj, [@{$pkeys}, $key], $objbase, $objscope, $deref, $sizeLim, $timeLim, $filter);
         push(@match_entries, @entries);
         $sizeLim = $sizeLim - @entries;
 
@@ -282,7 +287,7 @@ sub compare
         $retry++;
     }
 
-    my ($rc, $obj, $pkey) = $self->_getObject($dn);
+    my ($rc, $obj, $pkeys) = $self->_getObject($dn);
     if ($rc) {
         return $rc;
     }
@@ -291,7 +296,7 @@ sub compare
         my $entry;
         my $key;
 
-        ($rc, $key, $entry) = $self->_baseSearch($obj, $pkey, $dn, 0, 0, 1, 0, undef);
+        ($rc, $key, $entry) = $self->_baseSearch($obj, $pkeys, $dn, 0, 0, 1, 0, undef);
         if ($rc) {
             last DO;
         }
@@ -337,12 +342,12 @@ sub modify
         $retry++;
     }
 
-    my ($rc, $obj, $pkey) = $self->_getObject($dn);
+    my ($rc, $obj, $pkeys) = $self->_getObject($dn);
     if ($rc) {
         return $rc;
     }
 
-    return $self->_objModify($obj, $pkey, $dn, @list);
+    return $self->_objModify($obj, $pkeys, $dn, @list);
 }
 
 =pod
@@ -380,12 +385,12 @@ sub add
         $retry++;
     }
 
-    my ($rc, $obj, $pkey) = $self->_getObject($dn);
+    my ($rc, $obj, $pkeys) = $self->_getObject($dn);
     if ($rc) {
         return $rc;
     }
 
-    return $self->_objAdd($obj, $pkey, $dn, $entryStr);
+    return $self->_objAdd($obj, $pkeys, $dn, $entryStr);
 }
 
 =pod
@@ -460,28 +465,36 @@ sub delete
         $retry++;
     }
 
-    my ($rc, $obj, $pkey) = $self->_getObject($dn);
+    my ($rc, $obj, $pkeys) = $self->_getObject($dn);
     if ($rc) {
         return $rc;
     }
 
+    if ($obj->{conf}->{subcontainer}) {
+        my $level = split(/,/, ($dn =~ /^(.*),?$self->{suffix}$/i)[0]);
+        if ($obj->{conf}->{subcontainer}[0]->{level} == $level &&
+            $dn =~ /^$obj->{conf}->{subcontainer}[0]->{rdn}[0],/i) {
+            return LDAP_UNWILLING_TO_PERFORM;
+        }
+    }
+
     my $key;
 
-    ($rc, $key) = $self->_baseSearch($obj, $pkey, $dn, 0, 0, 1, 0, undef);
+    ($rc, $key) = $self->_baseSearch($obj, $pkeys, $dn, 0, 0, 1, 0, undef);
     if ($rc) {
         return $rc;
     }
 
     my @children;
     my $keys;
-    ($rc, $keys, @children) = $self->_childSearch($obj, $key, $dn, 1, 0, 1, 0, undef);
+    ($rc, $keys, @children) = $self->_childSearch($obj, [@{$pkeys}, $key], $dn, 1, 0, 1, 0, undef);
     if ($rc) {
         return $rc;
     } elsif (@children) {
         return LDAP_NOT_ALLOWED_ON_NONLEAF;
     }
 
-    return $self->_objDelete($obj, $pkey, $dn);
+    return $self->_objDelete($obj, $pkeys, $dn);
 }
 
 =pod
@@ -699,7 +712,7 @@ sub _checkConfig
         foreach my $lib (@{$conf->{libload}}) {
             eval "do \'$lib\'";
             if ($@) {
-                $self->log(level => 'alert', message => "storage do $lib: $@");
+                $self->log(level => 'alert', message => "Storage do $lib: $@");
                 return 1;
             }
         }
@@ -713,7 +726,7 @@ sub _checkConfig
     # multibyte character code
     if (defined($conf->{mbcode})) {
         if ($conf->{mbcode}[0] !~ /^(utf8|euc-jp|shiftjis|cp932)/) {
-            $self->log(level => 'alert', message => "character code is invalid");
+            $self->log(level => 'alert', message => "Character code is invalid");
             return 1;
         }
     } else {
@@ -750,15 +763,24 @@ sub _checkConfig
             my $oconf = $conf->{object}{$oname};
             my $entry;
 
-            if (!defined($oconf->{container})) {
-                $self->log(level => 'alert', message => "$oname object container doesn't exist");
-                return 1;
-            }
             $self->{object}{$oname}->{name} = $oname;
             $self->{object}{$oname}->{conf} = $oconf;
 
             # set container
-            if (defined($oconf->{container}[0]->{oname})) {
+            if (!defined($oconf->{container})) {
+                my $coname = 'top_'.$oname.'_container';
+                if (!defined($self->{object}{$coname})) {
+                    $self->{object}{$coname}->{entrystr} = '';
+                    # normalize suffix
+                    ($self->{object}{$coname}->{suffix} = $self->{suffix}) =~ tr/A-Z/a-z/;
+                }
+                # set child object
+                if (!defined($self->{object}{$coname}->{child})) {
+                    $self->{object}{$coname}->{child} = [];
+                }
+                push(@{$self->{object}{$coname}->{child}}, $self->{object}{$oname});
+                $self->{object}{$oname}->{parent} = $coname;
+            } elsif (defined($oconf->{container}[0]->{oname})) {
                 my $poname = $oconf->{container}[0]->{oname}[0];
                 $self->{object}{$oname}->{parent} = $poname;
                 if (!defined($self->{object}{$poname}->{child})) {
@@ -785,8 +807,12 @@ sub _checkConfig
                 $self->{object}{$oname}->{parent} = $coname;
             }
 
-            if (defined($oconf->{entry})) {
-                next;
+            # subcontainer
+            if (defined($oconf->{subcontainer})) {
+                if (!defined($oconf->{subcontainer}[0]->{rdn})) {
+                    $self->log(level => 'alert', message => "Set rdn in subcontainer");
+                    return 1;
+                }
             }
 
             foreach my $attr (keys(%{$oconf->{attr}})) {
@@ -799,13 +825,13 @@ sub _checkConfig
                 # multibyte character code
                 if (defined($oconf->{attr}{$attr}->{mbcode}) &&
                     !($oconf->{attr}{$attr}->{mbcode}[0] =~ /^(euc-jp|shiftjis)/)) {
-                    $self->log(level => 'alert', message => "character code is invalid");
+                    $self->log(level => 'alert', message => "Character code is invalid");
                     return 1;
                 }
             }
 
             if (!defined($oconf->{rdn}) || !defined($oconf->{attr}{$oconf->{rdn}[0]})) {
-                $self->log(level => 'alert', message => "rdn of $oname object is invalid");
+                $self->log(level => 'alert', message => "Rdn of $oname object is invalid");
                 return 1;
             }
 
@@ -813,7 +839,7 @@ sub _checkConfig
                 foreach my $si (@{$oconf->{strginfo}}) {
                     # check type of value
                     if (defined($si->{value}) && !ref($si->{value}[0])) {
-                        $self->log(level => 'alert', message => "type of strginfo value doesn't exist");
+                        $self->log(level => 'alert', message => "Type of strginfo value doesn't exist");
                         return 1;
                     }
                 }
@@ -829,7 +855,7 @@ sub _checkConfig
             }
 
             if ($self->_getTreeLevel($obj, 0) < 0) {
-                $self->log(level => 'alert', message => "depth of $oname subtree is too long");
+                $self->log(level => 'alert', message => "Depth of $oname subtree is too long");
                 return 1;
             }
         }
@@ -844,26 +870,26 @@ sub _getTreeLevel
     my ($obj, $current) = @_;
     my $maxlevel = 0;
 
-    $current++;
+    if (!defined($obj->{suffix}) || $obj->{entrystr}) {
+        $current++;
+    }
+
     if ($current > $maxLoopCount) {
         return -1;
     }
 
+    if (defined($obj->{conf}->{subcontainer})) {
+        $obj->{conf}->{subcontainer}[0]->{level} = $current;
+        $current++;
+    }
+
+    $obj->{level} = $current;
+
     if (!defined($obj->{child})) {
         return $current;
     }
 
     foreach my $child (@{$obj->{child}}) {
-        if (defined($child->{conf}->{entry})) {
-            my $entry;
-            my ($suffix) = ($obj->{entrystr} =~ /^dn: (.*)\n/);
-            if (!($entry = $self->buildEntryStr($suffix, $child->{conf}->{entry}[0]))) {
-                $self->log(level => 'alert', message => "container entry is invalid");
-                return -1;
-            }
-            $child->{entrystr} = $entry;
-        }
-
         my $level = $self->_getTreeLevel($child, $current);
         if ($level < 0) {
             return -1;
@@ -892,27 +918,31 @@ sub _getObject
     my $obj = undef;
 
     if (!defined($self->{object})) {
-        return (LDAP_UNWILLING_TO_PERFORM, $obj, undef);
+        return (LDAP_UNWILLING_TO_PERFORM, $obj, []);
     }
 
     foreach my $oname (keys %{$self->{object}}) {
-        $obj = $self->{object}{$oname};
+        my $tmpobj = $self->{object}{$oname};
+
+        if (defined($tmpobj->{suffix}) && $dn =~ /$tmpobj->{suffix}$/i) {
+            if (defined($obj) && length($obj->{suffix}) > length($tmpobj->{suffix})) {
+                next;
+            }
+            $obj = $tmpobj;
 
-        if (defined($obj->{suffix}) && $dn =~ /$obj->{suffix}$/i) {
             if ($dn =~ /^$obj->{suffix}$/i) {
-                return (LDAP_SUCCESS, $obj, undef);
-            } else {
-                last;
+                return (LDAP_SUCCESS, $obj, []);
             }
         }
     }
 
     if (!$obj) {
-        return (LDAP_NO_SUCH_OBJECT, $obj, undef);
+        return (LDAP_NO_SUCH_OBJECT, $obj, []);
     }
 
     my (@rdn_list) = split(/,/, ($dn =~ /^[^,]+,(.*),?$obj->{suffix}$/i)[0]);
     my @objs = ($obj);
+    my @pkeyarr = ();
     my $key;
     for (my $base = $obj->{suffix}; defined($obj->{child});
         @objs = @{$obj->{child}}, $base = pop(@rdn_list).','.$base) {
@@ -920,15 +950,29 @@ sub _getObject
         my $entry;
 
         for (my $i = 0; $i < @objs; $i++) {
-            ($rc, $key, $entry) = $self->_baseSearch($objs[$i], $key, $base, 0, 0, 1, 0, undef);
+            if (defined($objs[$i]->{suffix}) && !$objs[$i]->{entrystr}) {
+                $entry = "top_container";
+                $obj = $objs[$i];
+                last;
+            }
+            if (defined($objs[$i]->{conf}->{subcontainer})) {
+                my $rdn = $objs[$i]->{conf}->{subcontainer}[0]->{rdn}[0];
+                if ($base =~ /^$rdn,/i) {
+                    $key = $pkeyarr[$#pkeyarr];
+                    $entry = "subcontainer";
+                    last;
+                }
+            }
+
+            ($rc, $key, $entry) = $self->_baseSearch($objs[$i], \@pkeyarr, $base, 0, 0, 1, 0, undef);
             if ($entry) {
                 $obj = $objs[$i];
                 last;
             }
         }
-
+        push(@pkeyarr, $key);
         if (!$entry) {
-            return (LDAP_NO_SUCH_OBJECT, $obj, undef);
+            return (LDAP_NO_SUCH_OBJECT, $obj, []);
         }
         if (!@rdn_list) {
             last;
@@ -937,14 +981,20 @@ sub _getObject
 
     if (defined($obj->{child})) {
         foreach my $child (@{$obj->{child}}) {
-            if ((defined($child->{conf}->{rdn}) && $dn =~ /^$child->{conf}->{rdn}[0]=/i) ||
-                (defined($child->{entrystr}) && $child->{entrystr} =~ /^dn: $dn/i)) {
-                return (LDAP_SUCCESS, $child, $key);
+            if (defined($child->{conf}->{subcontainer}) &&
+                ($dn =~ /^$child->{conf}->{subcontainer}[0]->{rdn}[0],/i ||
+                    $dn =~ /^$child->{conf}->{rdn}[0]=[^,]+,$child->{conf}->{subcontainer}[0]->{rdn}[0]/i)) {
+                return (LDAP_SUCCESS, $child, \@pkeyarr);
+            }
+        }
+        foreach my $child (@{$obj->{child}}) {
+            if (defined($child->{conf}->{rdn}) && $dn =~ /^$child->{conf}->{rdn}[0]=/i) {
+                return (LDAP_SUCCESS, $child, \@pkeyarr);
             }
         }
     }
 
-    return (LDAP_NO_SUCH_OBJECT, $obj, undef);
+    return (LDAP_NO_SUCH_OBJECT, $obj, []);
 }
 
 sub _searchObjects
@@ -962,16 +1012,16 @@ sub _searchObjects
         if ($scope != 0) {
             foreach my $oname (keys %{$self->{object}}) {
                 if (defined($self->{object}{$oname}->{suffix})) {
-                    push(@objs, [$self->{object}{$oname}, undef]);
+                    push(@objs, [$self->{object}{$oname}, []]);
                 }
             }
         }
     } else {
-        my ($rc, $obj, $pkey) = $self->_getObject($base);
+        my ($rc, $obj, $pkeys) = $self->_getObject($base);
         if ($rc) {
             return ($rc, @objs);
         } else {
-            push(@objs, [$obj, $pkey]);
+            push(@objs, [$obj, $pkeys]);
         }
     }
 
@@ -981,15 +1031,38 @@ sub _searchObjects
 sub _baseSearch
 {
     my $self = shift;
-    my ($obj, $pkey, $base, $scope, $deref, $sizeLim, $timeLim, $filter) = @_;
+    my ($obj, $pkeys, $base, $scope, $deref, $sizeLim, $timeLim, $filter) = @_;
+    my $oconf = $obj->{conf};
+    my @pkeyarr = @{$pkeys};
 
-    if ($base !~ /$obj->{suffix}$/i) {
+    if (defined($obj->{suffix}) && $base !~ /$obj->{suffix}$/i) {
         return (LDAP_NO_SUCH_OBJECT, undef, undef);
     }
 
+    if (defined($oconf->{subcontainer})) {
+        my $level = split(/,/, ($base =~ /^(.*),?$self->{suffix}$/i)[0]);
+
+        if ($oconf->{subcontainer}[0]->{level} == $level) {
+            if ($base !~ /^$oconf->{subcontainer}[0]->{rdn}[0],/i) {
+                return (LDAP_NO_SUCH_OBJECT, undef, undef);
+            }
+
+            my $suffix = $base;
+            $suffix =~ s/^[^,]+,//;
+            my $entry = $self->buildEntryStr($suffix, $oconf->{subcontainer}[0]);
+            if ($self->parseFilter($filter, $entry) && $scope != 1 && $sizeLim) {
+                return (LDAP_SUCCESS, $pkeyarr[$#pkeyarr], $entry);
+            } else {
+                return (LDAP_SUCCESS, $pkeyarr[$#pkeyarr], undef);
+            }
+        }
+    }
+
     if (defined($obj->{entrystr})) {
         my $entry = $obj->{entrystr};
-        if ($entry !~ /^dn: $base\n/i) {
+        if (!$entry) {
+            return (LDAP_SUCCESS, undef, undef);
+        } elsif ($entry !~ /^dn: $base\n/i) {
             return (LDAP_NO_SUCH_OBJECT, undef, undef);
         } elsif ($self->parseFilter($filter, $entry) && $scope != 1 && $sizeLim) {
             return (LDAP_SUCCESS, undef, $entry);
@@ -1005,7 +1078,7 @@ sub _baseSearch
     }
     my $basefilter = Net::LDAP::Filter->new($filterStr);
 
-    my ($rc, $keys, @entries) = $self->_objSearch($obj, $pkey, $pdn, -1, $basefilter);
+    my ($rc, $keys, @entries) = $self->_objSearch($obj, $pkeys, $pdn, -1, $basefilter);
     if ($rc) {
         return ($rc, undef, undef);
     }
@@ -1016,52 +1089,72 @@ sub _baseSearch
 sub _childSearch
 {
     my $self = shift;
-    my ($obj, $pkey, $base, $scope, $deref, $sizeLim, $timeLim, $filter) = @_;
+    my ($obj, $pkeys, $base, $scope, $deref, $sizeLim, $timeLim, $filter) = @_;
+    my $oconf = $obj->{conf};
     my @match_entries = ();
     my @match_keys = ();
     my $rc = LDAP_SUCCESS;
+    my @pkeyarr = @{$pkeys};
+    my @objs;
 
-    if ($scope == 0 || !defined($obj->{child}) || !$pkey && !defined($obj->{entrystr})) {
+    if ($scope == 0) {
         return ($rc, \@match_keys, ());
     }
 
     my @entries;
     my $keys;
-    for (my $i = 0; $i < @{$obj->{child}}; $i++) {
+    my $level = split(/,/, ($base =~ /^(.*),?$self->{suffix}$/i)[0]);
+
+    if (defined($oconf->{subcontainer}) && $oconf->{subcontainer}[0]->{level} == $level) {
+        @objs = ($obj);
+    } else {
+        if (!defined($obj->{child})) {
+            return ($rc, \@match_keys, @match_entries);
+        }
+        @objs = @{$obj->{child}};
+    }
+
+    for (my $i = 0; $i < @objs; $i++) {
+        my $cbase;
         my @children = ();
         my $ckeys;
 
-        if (defined($obj->{child}[$i]->{entrystr})) {
-            my $entry = $obj->{child}[$i]->{entrystr};
-            if ($self->parseFilter($filter, $entry) && $scope != 0 && $sizeLim) {
+        if (defined($objs[$i]->{conf}->{subcontainer}) &&
+            $objs[$i]->{conf}->{subcontainer}[0]->{level} != $level) {
+            my $entry = $self->buildEntryStr($base, $objs[$i]->{conf}->{subcontainer}[0]);
+            if ($self->parseFilter($filter, $entry) && $sizeLim) {
                 push(@match_entries, $entry);
                 push(@match_keys, undef);
             }
+            if ($scope == 1) {
+                next;
+            }
+            $cbase = $objs[$i]->{conf}->{subcontainer}[0]->{rdn}[0].",$base";
         } else {
-            ($rc, $keys, @entries) = $self->_objSearch($obj->{child}[$i], $pkey, $base, $sizeLim, $filter);
-            push(@match_entries, @entries);
-            push(@match_keys, @{$keys});
+            $cbase = $base;
         }
 
-        if ($scope == 1 || !defined($obj->{child}[$i]->{child}) || $rc) {
+        ($rc, $keys, @entries) = $self->_objSearch($objs[$i], $pkeys, $cbase, $sizeLim, $filter);
+        push(@match_entries, @entries);
+        push(@match_keys, @{$keys});
+
+        if ($scope == 1 || !defined($objs[$i]->{child}) || $rc) {
             next;
         }
         $sizeLim = $sizeLim - @match_entries;
 
-        if (defined($obj->{child}[$i]->{entrystr})) {
-            push(@children, $obj->{child}[$i]->{entrystr});
-            my @tmp_keys = (undef);
-            $ckeys = \@tmp_keys;
-        } else {
-            ($rc, $ckeys, @children) = $self->_objSearch($obj->{child}[$i], $pkey, $base, -1, undef);
-            if ($rc) {
-                return ($rc, \@match_keys, @match_entries);
-            }
+        ($rc, $ckeys, @children) = $self->_objSearch($objs[$i], $pkeys, $cbase, -1, undef);
+        if ($rc) {
+            return ($rc, \@match_keys, @match_entries);
         }
 
         for (my $j = 0; $j < @{$ckeys}; $j++) {
             my ($childdn) = ($children[$j] =~ /^dn: (.*)\n/);
-            ($rc, $keys, @entries) = $self->_childSearch($obj->{child}[$i], ${$ckeys}[$j], $childdn, $scope, $deref, $sizeLim, $timeLim, $filter);
+            if (!$childdn) {
+                return (LDAP_OTHER, \@match_keys, @match_entries);
+            }
+
+            ($rc, $keys, @entries) = $self->_childSearch($objs[$i], [@pkeyarr, ${$ckeys}[$j]], $childdn, $scope, $deref, $sizeLim, $timeLim, $filter);
             if ($rc) {
                 return ($rc, \@match_keys, @match_entries);
             }
@@ -1124,6 +1217,10 @@ sub _getParentDn
         return undef;
     }
 
+    if (defined($obj->{conf}->{subcontainer})) {
+        $prdn = "$obj->{conf}->{subcontainer}[0]->{rdn}[0],$prdn";
+    }
+
     return "$prdn,$ppdn";
 }
 
