@@ -115,7 +115,7 @@ sub _checkConfig
 
 =pod
 
-=head2 _objSearch($obj, $pkey, $suffix, $sizeLim, $timeLim, $filter)
+=head2 _objSearch($obj, $pkeys, $suffix, $sizeLim, $timeLim, $filter)
 
 Search the appropriate records in the object's file.
 
@@ -124,9 +124,10 @@ Search the appropriate records in the object's file.
 sub _objSearch
 {
     my $self = shift;
-    my ($obj, $pkey, $suffix, $sizeLim, $filter) = @_;
+    my ($obj, $pkeys, $suffix, $sizeLim, $filter) = @_;
     my $conf = $self->{_config};
     my $oconf = $obj->{conf};
+    my $pkey = ${$pkeys}[$#{$pkeys}];
     my $dlm = $conf->{delim}[0];
     my $valdlm = $conf->{valdelim}[0];
     my @match_entries = ();
@@ -220,7 +221,7 @@ sub _objSearch
 
 =pod
 
-=head2 _objModify($obj, $pkey, $dn, @list)
+=head2 _objModify($obj, $pkeys, $dn, @list)
 
 Write the modified record to the temporary file.
 
@@ -229,9 +230,10 @@ Write the modified record to the temporary file.
 sub _objModify
 {
     my $self = shift;
-    my ($obj, $pkey, $dn, @list) = @_;
+    my ($obj, $pkeys, $dn, @list) = @_;
     my $conf = $self->{_config};
     my $oconf = $obj->{conf};
+    my $pkey = ${$pkeys}[$#{$pkeys}];
     my $dlm = $conf->{delim}[0];
     my $valdlm = $conf->{valdelim}[0];
     my $match = 0;
@@ -407,7 +409,7 @@ sub _objModify
 
 =pod
 
-=head2 _objAdd($obj, $pkey, $dn, $entryStr)
+=head2 _objAdd($obj, $pkeys, $dn, $entryStr)
 
 Copy the object's file to the temporary file and add the record to it.
 
@@ -416,9 +418,10 @@ Copy the object's file to the temporary file and add the record to it.
 sub _objAdd
 {
     my $self = shift;
-    my ($obj, $pkey, $dn,  $entryStr) = @_;
+    my ($obj, $pkeys, $dn,  $entryStr) = @_;
     my $conf = $self->{_config};
     my $oconf = $obj->{conf};
+    my $pkey = ${$pkeys}[$#{$pkeys}];
     my $dlm = $conf->{delim}[0];
     my $valdlm = $conf->{valdelim}[0];
     my $rc = LDAP_SUCCESS;
@@ -551,7 +554,7 @@ sub _objAdd
 
 =pod
 
-=head2 _objDelete($dn)
+=head2 _objDelete($obj, $pkeys, $dn)
 
 Copy the object's file from which the appropriate record is deleted to the temporary file.
 
@@ -560,9 +563,10 @@ Copy the object's file from which the appropriate record is deleted to the tempo
 sub _objDelete
 {
     my $self = shift;
-    my ($obj, $pkey, $dn) = @_;
+    my ($obj, $pkeys, $dn) = @_;
     my $conf = $self->{_config};
     my $oconf = $obj->{conf};
+    my $pkey = ${$pkeys}[$#{$pkeys}];
     my $dlm = $conf->{delim}[0];
     my $valdlm = $conf->{valdelim}[0];
     my $rc = LDAP_NO_SUCH_OBJECT;
@@ -821,14 +825,14 @@ sub _getAttrKeys
         for (my $i = 0; $i < @values && $values[$i]; $i++) {
             my $aobj;
             my $attrkey;
-            my $attrpkey;
+            my $attrpkeys;
 
-            ($rc, $aobj, $attrpkey) = $self->_getObject($values[$i]);
+            ($rc, $aobj, $attrpkeys) = $self->_getObject($values[$i]);
             if ($rc) {
                 return (-1, ());
             }
 
-            ($rc, $attrkey) =$self->_baseSearch($aobj, $attrpkey, $values[$i], 0, 0, 1, 0, undef, 0, ('dn'));
+            ($rc, $attrkey) =$self->_baseSearch($aobj, $attrpkeys, $values[$i], 0, 0, 1, 0, undef, 0, ('dn'));
             if ($rc || !$attrkey) {
                 return (-1, ());
             }
