@@ -180,7 +180,7 @@ sub modify
             push(@values, shift @list);
         }
 
-        if ($key =~ /^modifyTimestamp$/i) {
+        if ($key =~ /^(modifyTimestamp|plainPassword)$/i) {
             next;
         }
 
@@ -249,7 +249,7 @@ sub add
     my @info = split(/\n/, $entryStr);
     foreach my $attr (@info) {
         my ($key, $val) = split(/: /, $attr);
-        if ($key =~ /^(createTimestamp|modifyTimestamp)$/i) {
+        if ($key =~ /^(createTimestamp|modifyTimestamp|plainpassword)$/i) {
             next;
         }
 
