@@ -129,26 +129,36 @@ sub hashPasswd
 sub _getConnect
 {
     my $self = shift;
+
+    return $self->_connect(\$self->{db});
+}
+
+sub _connect
+{
+    my $self = shift;
+    my ($db) = @_;
     my $conf = $self->{_config};
 
-    if (defined($self->{db})) {
-        my $sth = $self->{db}->prepare("select 'check';");
+    if (defined(${$db})) {
+        my $sth = ${$db}->prepare("select 'check';");
         if (!$sth->execute) {
             $self->log(level => 'err', message => "Connection check failed");
+            ${$db}->disconnect();
+            undef(${$db});
         } else {
             return 0;
         }
         $sth->finish;
     }
 
-    $self->{db} = DBI->connect($conf->{dsn}[0], $conf->{admin}[0], $conf->{passwd}[0]);
-    if (!$self->{db}) {
+    ${$db} = DBI->connect($conf->{dsn}[0], $conf->{admin}[0], $conf->{passwd}[0]);
+    if (!${$db}) {
         $self->log(level => 'alert', message => "Can't connect $conf->{dsn}[0]: ".$DBI::errstr);
         return -1;
     }
 
-    if (defined($conf->{initquery}) && !$self->{db}->do($conf->{initquery}[0])) {
-        $self->log(level => 'crit', message => "$conf->{initquery}[0] failed: ".$self->{db}->errstr);
+    if (defined($conf->{initquery}) && !${$db}->do($conf->{initquery}[0])) {
+        $self->log(level => 'crit', message => "$conf->{initquery}[0] failed: ".${$db}->errstr);
         return -1;
     }
 
@@ -205,6 +215,21 @@ sub _checkConfig
                 }
             }
         }
+
+        # check paramset
+        foreach my $func (keys %{$oconf->{plugin}}) {
+            my $plugin = $oconf->{plugin}{$func};
+            foreach my $key1 (keys %{$plugin->{param}}) {
+                if (defined($plugin->{param}{$key1}->{type}) &&
+                    $plugin->{param}{$key1}->{type} eq 'paramset' &&
+                        defined($conf->{paramset}{$plugin->{param}{$key1}->{value}})) {
+                    my $paramset = $conf->{paramset}{$plugin->{param}{$key1}->{value}};
+                    foreach my $key2 (keys %{$paramset->{param}}) {
+                        $plugin->{param}{$key2} = $paramset->{param}{$key2};
+                    }
+                }
+            }
+        }
     }
 
     return 0;
@@ -212,7 +237,7 @@ sub _checkConfig
 
 =pod
 
-=head2 _objSearch($obj, $parent, $suffix, $sizeLim, $filter)
+=head2 _objSearch($obj, $pkeys, $suffix, $sizeLim, $filter)
 
 Search the appropriate records in the object's table.
 
@@ -221,9 +246,10 @@ Search the appropriate records in the object's table.
 sub _objSearch
 {
     my $self = shift;
-    my ($obj, $pkey, $suffix, $sizeLim, $filter) = @_;
+    my ($obj, $pkeys, $suffix, $sizeLim, $filter) = @_;
     my $conf = $self->{_config};
     my $oconf = $obj->{conf};
+    my $pkey = ${$pkeys}[$#{$pkeys}];
     my @match_entries = ();
     my @match_keys = ();
     my $rc = LDAP_SUCCESS;
@@ -251,14 +277,19 @@ sub _objSearch
         if ($pkey && defined($oconf->{container}) && !defined($oconf->{container}[0]->{rdn})) {
             my $poconf = $self->{object}{$oconf->{container}[0]->{oname}[0]}->{conf};
 
-            $from = "$from, $poconf->{table}[0]";
+            if ($oconf->{table}[0] ne $poconf->{table}[0]) {
+                $from = "$from, $poconf->{table}[0]";
+                $where = "$where and $poconf->{table}[0].$poconf->{id}[0]->{column}[0] = \'$pkey\'";
+            }
+
             if (defined($oconf->{container}[0]->{fromtbls})) {
                 $from = "$from, $oconf->{container}[0]->{fromtbls}[0]";
             }
 
-            $where = "$where and $poconf->{table}[0].$poconf->{id}[0]->{column}[0] = \'$pkey\'";
             if (defined($oconf->{container}[0]->{joinwhere})) {
-                $where = "$where and $oconf->{container}[0]->{joinwhere}[0]";
+                my $pwhere = $oconf->{container}[0]->{joinwhere}[0];
+                $pwhere =~ s/\%c/$pkey/g;
+                $where = "$where and $pwhere";
             }
         }
 
@@ -360,7 +391,7 @@ sub _objSearch
 
 =pod
 
-=head2 _objModify($obj, $peky, $dn, @list)
+=head2 _objModify($obj, $pekys, $dn, @list)
 
 Update the data of the object's table and insert the data to the attribute's table.
 
@@ -369,9 +400,10 @@ Update the data of the object's table and insert the data to the attribute's tab
 sub _objModify
 {
     my $self = shift;
-    my ($obj, $pkey, $dn, @list) = @_;
+    my ($obj, $pkeys, $dn, @list) = @_;
     my $conf = $self->{_config};
     my $oconf = $obj->{conf};
+    my $pkey = ${$pkeys}[$#{$pkeys}];
     my $rc = LDAP_SUCCESS;
 
     DO: {
@@ -385,7 +417,7 @@ sub _objModify
         my $entry;
         my $key;
 
-        ($rc, $key, $entry) = $self->_baseSearch($obj, $pkey, $dn, 0, 0, 1, 0, undef);
+        ($rc, $key, $entry) = $self->_baseSearch($obj, $pkeys, $dn, 0, 0, 1, 0, undef);
         if ($rc) {
             last DO;
         }
@@ -395,6 +427,29 @@ sub _objModify
             last DO;
         }
 
+        if (defined($oconf->{plugin})) {
+            foreach my $func (keys %{$oconf->{plugin}}) {
+                my $plugin = $oconf->{plugin}{$func};
+                my $error;
+                if (!defined($plugin->{op}) || $plugin->{op}[0] eq 'modify') {
+                    eval "(\$rc, \$error) = LISM::Storage::$func(\$self, \$oconf, \$plugin->{param}, \$pkeys, \$key, \$dn, \@list)";
+                    if ($@) {
+                        $self->log(level => 'err', message => "plugin $func failed in modify operation: $@");
+                        $rc = LDAP_OTHER;
+                        last DO;
+                    } elsif ($rc) {
+                        $self->log(level => 'err', message => "plugin $func returns error in modify operation: $error");
+                        $rc = LDAP_OTHER;
+                        last DO;
+                    }
+                }
+            }
+        }
+
+        if (defined($oconf->{noop}) && grep(/^modify$/i, @{$oconf->{noop}})) {
+            last DO;
+        }
+
         while ( @list > 0 ) {
             my $action = shift @list;
             my $attr    = lc(shift @list);
@@ -594,7 +649,7 @@ sub _objModify
 
 =pod
 
-=head2 _objAdd($obj, $pkey, $dn, $entryStr)
+=head2 _objAdd($obj, $pkeys, $dn, $entryStr)
 
 Insert the data to the object's and the attribute's table.
 
@@ -603,34 +658,52 @@ Insert the data to the object's and the attribute's table.
 sub _objAdd
 {
     my $self = shift;
-    my ($obj, $pkey, $dn, $entryStr) = @_;
+    my ($obj, $pkeys, $dn, $entryStr) = @_;
     my $conf = $self->{_config};
     my $oconf = $obj->{conf};
+    my $pkey = ${$pkeys}[$#{$pkeys}];
     my $rc = LDAP_SUCCESS;
 
+    # check whether the entry already exists
+    my $entry;
+    my $key;
+
+    ($rc, $key, $entry) = $self->_baseSearch($obj, $pkeys, $dn, 0, 0, 1, 0, undef);
+    if ($rc && $rc != LDAP_NO_SUCH_OBJECT) {
+        return $rc;
+    } elsif ($entry) {
+        return LDAP_ALREADY_EXISTS;
+    }
+
+    if (defined($oconf->{plugin})) {
+        foreach my $func (keys %{$oconf->{plugin}}) {
+            my $plugin = $oconf->{plugin}{$func};
+            my $error;
+            if (!defined($plugin->{op}) || $plugin->{op}[0] eq 'add') {
+                eval "(\$rc, \$error) = LISM::Storage::$func(\$self, \$oconf, \$plugin->{param}, \$pkeys, \$dn, \$entryStr)";
+                if ($@) {
+                    $self->log(level => 'err', message => "plugin $func failed in add operation: $@");
+                    return LDAP_OTHER;
+                } elsif ($rc) {
+                    $self->log(level => 'err', message => "plugin $func returns error in add operation: $error");
+                    return LDAP_OTHER;
+                }
+            }
+        }
+    }
+
     DO: {
         # start transaction
         if (!$self->{db}->begin_work) {
             $self->log(level => 'err', message => "Can't begin: ".$self->{db}->errstr);
-            $rc = LDAP_OPERATIONS_ERROR;
-            last DO;
-        }
-
-        # check whether the entry already exists
-        my $entry;
-        my $key;
-
-        ($rc, $key, $entry) = $self->_baseSearch($obj, $pkey, $dn, 0, 0, 1, 0, undef);
-        if ($rc && $rc != LDAP_NO_SUCH_OBJECT) {
-             last DO;
-        } elsif ($entry) {
-             $rc = LDAP_ALREADY_EXISTS;
-             last DO;
+            return LDAP_OPERATIONS_ERROR;
         }
 
         my @cols;
         my @values;
         my %attrs;
+        my $sql;
+        my $sth;
 
         if (defined($oconf->{id}[0]->{sequence})) {
             push(@cols, $oconf->{id}[0]->{column}[0]);
@@ -676,25 +749,27 @@ sub _objAdd
             }
         }
 
-        my $sql = "insert into $oconf->{table}[0](".join(', ', @cols).") values(".join(', ', @values).")";
+        if (!defined($oconf->{noop}) || !grep(/^add$/i, @{$oconf->{noop}})) {
+            $sql = "insert into $oconf->{table}[0](".join(', ', @cols).") values(".join(', ', @values).")";
 
-        # multibyte value
-        $sql = encode($conf->{mbcode}[0], $sql);
+            # multibyte value
+            $sql = encode($conf->{mbcode}[0], $sql);
 
-        my $sth = $self->{db}->prepare("$sql");
-        if (!$sth->execute) {
-            $self->log(level => 'err', message => "Adding entry by \"$sql\" failed: ".$sth->errstr);
-            $rc = LDAP_OPERATIONS_ERROR;
-            last DO;
+            $sth = $self->{db}->prepare("$sql");
+            if (!$sth->execute) {
+                $self->log(level => 'err', message => "Adding entry by \"$sql\" failed: ".$sth->errstr);
+                $rc = LDAP_OPERATIONS_ERROR;
+                last DO;
+            }
+            $sth->finish;
         }
-        $sth->finish;
 
         # get the added object's id from the table
         my ($rdn, $pdn) = ($dn=~ /^([^,]+),(.*)$/);
         my $filter = Net::LDAP::Filter->new("(".encode('utf8', $rdn).")");
         my $keys;
 
-        ($rc, $keys) = $self->_objSearch($obj, '', $pdn, -1, $filter);
+        ($rc, $keys) = $self->_objSearch($obj, [], $pdn, -1, $filter);
         if ($rc || !@{$keys}) {
             $self->log(level => 'err', message => "Can't get id of $dn from the table");
             $rc = LDAP_OPERATIONS_ERROR;
@@ -732,13 +807,15 @@ sub _objAdd
             }
         }
 
-        # add the values in the attribute's table
-        foreach my $attr (keys %attrs) {
-            foreach my $value (@{$attrs{$attr}}) {
-                if ($rc = $self->_addAttrValues($oconf, $key, $pkey, $attr, $value, $dn, $entryStr)) {
-                    $self->log(level => 'err', message => "Adding values of $attr failed");
-                    $rc = LDAP_OPERATIONS_ERROR;
-                    last DO;
+        if (!defined($oconf->{noop}) || !grep(/^add$/i, @{$oconf->{noop}})) {
+            # add the values in the attribute's table
+            foreach my $attr (keys %attrs) {
+                foreach my $value (@{$attrs{$attr}}) {
+                    if ($rc = $self->_addAttrValues($oconf, $key, $pkey, $attr, $value, $dn, $entryStr)) {
+                        $self->log(level => 'err', message => "Adding values of $attr failed");
+                        $rc = LDAP_OPERATIONS_ERROR;
+                        last DO;
+                    }
                 }
             }
         }
@@ -768,7 +845,7 @@ sub _objAdd
 
 =pod
 
-=head2 _objDelete
+=head2 _objDelete($obj, $pkeys, $dn)
 
 delete the data from the object's and the attribute's table.
 
@@ -777,9 +854,10 @@ delete the data from the object's and the attribute's table.
 sub _objDelete
 {
     my $self = shift;
-    my ($obj, $pkey, $dn) = @_;
+    my ($obj, $pkeys, $dn) = @_;
     my $conf = $self->{_config};
     my $oconf = $obj->{conf};
+    my $pkey = ${$pkeys}[$#{$pkeys}];
     my $rc = LDAP_SUCCESS;
 
     DO: {
@@ -793,7 +871,7 @@ sub _objDelete
         # get the object's id from the table
         my $key;
 
-        ($rc, $key) = $self->_baseSearch($obj, $pkey, $dn, 0, 0, 1, 0, undef);
+        ($rc, $key) = $self->_baseSearch($obj, $pkeys, $dn, 0, 0, 1, 0, undef);
         if ($rc) {
             last DO;
         }
@@ -805,7 +883,7 @@ sub _objDelete
         my $sql;
         my $sth;
 
-        if (!defined($conf->{nodelete}) || $conf->{nodelete} ne 'on') {
+        if (!defined($oconf->{noop}) || !grep(/^delete$/i, @{$oconf->{noop}})) {
             # delete the values from the attribute's table
             foreach my $attr (keys %{$oconf->{attr}}) {
                 if (defined($oconf->{attr}{$attr}) && defined($oconf->{attr}{$attr}->{delproc})) {
@@ -842,34 +920,51 @@ sub _objDelete
             }
         }
 
-        if (defined($conf->{nodelete}) && $conf->{nodelete} eq 'on') {
-            last DO;
-        }
+        if (!defined($oconf->{noop}) || !grep(/^delete$/i, @{$oconf->{noop}})) {
+            # delete the link with container
+            if (defined($oconf->{container}) && defined($oconf->{container}[0]->{delproc})) {
+                $sql = $oconf->{container}[0]->{delproc}[0];
+                $sql =~ s/\%o/$key/g;
+                $sql =~ s/\%c/$pkey/g;
+                $sql = $self->_funcParse($sql, $dn);
+                $sth = $self->{db}->prepare("$sql");
+                if (!$sth->execute) {
+                    $self->log(level => 'err', message => "Adding link of container by \"$sql\" failed: ".$sth->errstr);
+                    $rc = LDAP_OPERATIONS_ERROR;
+                    last DO;
+                }
+                $sth->finish;
+            }
 
-        # delete the link with container
-        if (defined($oconf->{container}) && defined($oconf->{container}[0]->{delproc})) {
-            $sql = $oconf->{container}[0]->{delproc}[0];
-            $sql =~ s/\%o/$key/g;
-            $sql =~ s/\%c/$pkey/g;
-            $sql = $self->_funcParse($sql, $dn);
-            $sth = $self->{db}->prepare("$sql");
+            # delete the appropriate record from the object's table
+            $sql = "delete from $oconf->{table}[0] where $oconf->{id}[0]->{column}[0] = \'$key\'";
+            $sth = $self->{db}->prepare($sql);
             if (!$sth->execute) {
-                $self->log(level => 'err', message => "Adding link of container by \"$sql\" failed: ".$sth->errstr);
+                $self->log(level => 'err', message => "Deleting entry by \"$sql\" failed: ".$sth->errstr);
                 $rc = LDAP_OPERATIONS_ERROR;
                 last DO;
             }
             $sth->finish;
         }
 
-        # delete the appropriate record from the object's table
-        $sql = "delete from $oconf->{table}[0] where $oconf->{id}[0]->{column}[0] = \'$key\'";
-        $sth = $self->{db}->prepare($sql);
-        if (!$sth->execute) {
-            $self->log(level => 'err', message => "Deleting entry by \"$sql\" failed: ".$sth->errstr);
-            $rc = LDAP_OPERATIONS_ERROR;
-            last DO;
+        if (defined($oconf->{plugin})) {
+            foreach my $func (keys %{$oconf->{plugin}}) {
+                my $plugin = $oconf->{plugin}{$func};
+                my $error;
+                if (!defined($plugin->{op}) || $plugin->{op}[0] eq 'delete') {
+                    eval "(\$rc, \$error) = LISM::Storage::$func(\$self, \$oconf, \$plugin->{param}, \$pkeys, \$key, \$dn)";
+                    if ($@) {
+                        $self->log(level => 'err', message => "plugin $func failed in delete operation: $@");
+                        $rc = LDAP_OTHER;
+                        last DO;
+                    } elsif ($rc) {
+                        $self->log(level => 'err', message => "plugin $func returns error in delete operation: $error");
+                        $rc = LDAP_OTHER;
+                        last DO;
+                    }
+                }
+            }
         }
-        $sth->finish;
     }
 
     if ($rc) {
@@ -904,14 +999,22 @@ sub _getParentRdn
 
     $selexpr = "$poconf->{table}[0].$poconf->{id}[0]->{column}[0],$poconf->{table}[0].$poconf->{attr}{$poconf->{rdn}[0]}->{column}[0]";
 
-    $from = "$oconf->{table}[0], $poconf->{table}[0]";
+    $from = $oconf->{table}[0];
+    if ($from ne $poconf->{table}[0]) {
+        $from = "$from,$poconf->{table}[0]";
+    }
     if (defined($oconf->{container}[0]->{fromtbls})) {
         $from = "$from,$oconf->{container}[0]->{fromtbls}[0]";
     }
 
     $where = "$oconf->{table}[0].$oconf->{id}[0]->{column}[0] = \'$key\'";
     if (defined($oconf->{container}[0]->{joinwhere})) {
-        $where = "$where and $oconf->{container}[0]->{joinwhere}[0]";
+        if ($oconf->{container}[0]->{joinwhere}[0] =~ /\%c/) {
+            my ($pidcol) = ($oconf->{container}[0]->{joinwhere}[0] =~ /([^ ]+) *= *\%c/);
+            $where = "$poconf->{table}[0].$poconf->{id}[0]->{column}[0] = (select $pidcol from $oconf->{table}[0] where $where)";
+        } else {
+            $where = "$where and $oconf->{container}[0]->{joinwhere}[0]";
+        }
     }
 
     my $sql = "select $selexpr from $from where $where";
@@ -1046,13 +1149,13 @@ sub _addAttrValues
         $sql =~ s/\%o/$key/g;
         $sql =~ s/\%p/$pkey/g;
         if (defined($oconf->{attr}{$attr}->{oname})) {
-            my ($rc, $aobj, $attrpkey) = $self->_getObject($value);
+            my ($rc, $aobj, $attrpkeys) = $self->_getObject($value);
             if ($rc) {
                 return -1;
             }
 
             my $attrkey;
-            ($rc, $attrkey) =$self->_baseSearch($aobj, $attrpkey, $value, 0, 0, 1, 0, undef);
+            ($rc, $attrkey) =$self->_baseSearch($aobj, $attrpkeys, $value, 0, 0, 1, 0, undef);
             if ($rc) {
                 $self->log(level => 'err', message => "Can't get id of $value in the table");
                 return -1;
@@ -1066,13 +1169,13 @@ sub _addAttrValues
                 $aobj->{conf}->{parentrel}[0] eq 'tight') {
                 my $contdn;
                 ($contdn = $value) =~ s/^[^,]+,//;
-                my ($rc, $cobj, $contpkey) = $self->_getObject($contdn);
+                my ($rc, $cobj, $contpkeys) = $self->_getObject($contdn);
                 if ($rc) {
                     return -1;
                 }
 
                 my $contkey;
-                ($rc, $contkey) =$self->_baseSearch($cobj, $contpkey, $contdn, 0, 0, 1, 0, undef);
+                ($rc, $contkey) =$self->_baseSearch($cobj, $contpkeys, $contdn, 0, 0, 1, 0, undef);
                 if ($rc) {
                     $self->log(level => 'err', message => "Can't get id of $contdn in the table");
                     return -1;
@@ -1119,13 +1222,13 @@ sub _delAttrValues
         $sql =~ s/\%p/$pkey/g;
         if ($value) {
             if (defined($oconf->{attr}{$attr}->{oname})) {
-                my ($rc, $aobj, $attrpkey) = $self->_getObject($value);
+                my ($rc, $aobj, $attrpkeys) = $self->_getObject($value);
                 if ($rc) {
                     return -1;
                 }
 
                 my $attrkey;
-                ($rc, $attrkey) =$self->_baseSearch($aobj, $attrpkey, $value, 0, 0, 1, 0, undef);
+                ($rc, $attrkey) =$self->_baseSearch($aobj, $attrpkeys, $value, 0, 0, 1, 0, undef);
                 if ($rc) {
 		    $self->log(level => 'err', message => "Can't get id of $value from the table");
                     return -1;
@@ -1139,13 +1242,13 @@ sub _delAttrValues
                     $aobj->{conf}->{parentrel}[0] eq 'tight') {
                     my $contdn;
                     ($contdn = $value) =~ s/^[^,]+,//;
-                    my ($rc, $cobj, $contpkey) = $self->_getObject($contdn);
+                    my ($rc, $cobj, $contpkeys) = $self->_getObject($contdn);
                     if ($rc) {
                         return -1;
                     }
 
                     my $contkey;
-                    ($rc, $contkey) =$self->_baseSearch($cobj, $contpkey, $contdn, 0, 0, 1, 0, undef);
+                    ($rc, $contkey) =$self->_baseSearch($cobj, $contpkeys, $contdn, 0, 0, 1, 0, undef);
                     if ($rc) {
                         $self->log(level => 'err', message => "Can't get id of $contdn in the table");
                         return -1;
@@ -1268,17 +1371,21 @@ sub _filter2sql
     } elsif (defined($oconf->{attr}{$attr}->{column})) {
         ${$where} = "lower($oconf->{table}[0].$oconf->{attr}{$attr}->{column}[0]) $type lower(\'$value\')";
     } elsif (defined($oconf->{attr}{$attr}->{oname})) {
-        my $attrobj = $conf->{object}{$oconf->{attr}{$attr}->{oname}[0]};
-        my $aconf = $attrobj->{conf};
+        my $aconf = $self->{object}{$oconf->{attr}{$attr}->{oname}[0]}->{conf};
+
+        if (!grep(/^$aconf->{table}[0]$/, split(/, */, ${$from}))) {
+            ${$from} = "${$from},$aconf->{table}[0]";
+        }
         if (defined($oconf->{attr}{$attr}->{fromtbls})) {
             ${$from} = "${$from},$oconf->{attr}{$attr}->{fromtbls}[0]";
         }
 
-        ($value) = ($value =~ /^$attrobj->{rdn}[0]=([^,]*)/i);
-        ${$where} = "lower($attrobj->{table}[0].$attrobj->{attr}{$attrobj->{rdn}[0]}->{column}[0]) $type lower(\'$value\')";
+        ($value) = ($value =~ /^$aconf->{rdn}[0]=([^,]*)/i);
+        my $subquery = "select $oconf->{table}[0].$oconf->{id}[0]->{column}[0] from ${$from} where lower($aconf->{table}[0].$aconf->{attr}{$aconf->{rdn}[0]}->{column}[0]) $type lower(\'$value\')";
         if (defined($oconf->{attr}{$attr}->{joinwhere})) {
-            ${$where} = "${$where} and $oconf->{attr}{$attr}->{joinwhere}[0]";
+            $subquery = "$subquery and $oconf->{attr}{$attr}->{joinwhere}[0]";
         }
+        ${$where} = "$oconf->{table}[0].$oconf->{id}[0]->{column}[0] in ($subquery)";
 
         if (defined($aconf->{container}) && !defined($aconf->{container}[0]->{rdn})) {
             my $paconf = $self->{object}{$aconf->{container}[0]->{oname}[0]}->{conf};
@@ -1296,10 +1403,12 @@ sub _filter2sql
         if (defined($oconf->{attr}{$attr}->{fromtbls})) {
             ${$from} = "${$from},$oconf->{attr}{$attr}->{fromtbls}[0]";
         }
-        ${$where} = "lower($oconf->{attr}{$attr}->{selexpr}[0]) $type lower(\'$value\')";
+
+        my $subquery =  "select $oconf->{table}[0].$oconf->{id}[0]->{column}[0] from ${$from} where lower($oconf->{attr}{$attr}->{selexpr}[0]) $type lower(\'$value\')";
         if (defined($oconf->{attr}{$attr}->{joinwhere})) {
-            ${$where} = "${$where} and $oconf->{attr}{$attr}->{joinwhere}[0]";
+            $subquery = "$subquery and $oconf->{attr}{$attr}->{joinwhere}[0]";
         }
+        ${$where} = "$oconf->{table}[0].$oconf->{id}[0]->{column}[0] in ($subquery)";
     }
 
     return 1;
